/* 
 * This code solves the following Poisson's equation with zero Dirichlet BC:
 * 	 - \div (a(x) \grad u(x)) = f(x)
 *
 * Note: to extract DOFs and errors from the output:
 *	awk '/DOF/ {n=$2} /Error/ {print n, $4}'
 */

#include "phg.h"

#include <string.h>
#include <math.h>


static void
build_linear_system(SOLVER *solver, DOF *u_h, DOF *a, DOF *f)
{
    GRID *g = u_h->g;
    ELEMENT *e;

    assert(u_h->dim == 1);
    ForAllElements(g, e) {	/* loop on elements */
	int N = DofGetNBas(u_h, e);	/* number of bases in the element */
	int ii, jj;
	INT i, j;
	FLOAT mat, rhs, buf[N];

	for (ii = 0; ii < N; ii++) {	/* loop on bases in current element */
	    i = phgSolverMapE2L(solver, 0, e, ii);
	    if (phgDofDirichletBC(u_h, e, ii, NULL, buf, &rhs, DOF_PROJ_NONE)) {
		/* current node is on the Dirichlet boundary */
		for (jj = 0; jj < N; jj++) {
		    j = phgSolverMapE2L(solver, 0, e, jj);
		    phgSolverAddMatrixEntry(solver, i, j, buf[jj]);
		}
	    }
	    else {
		/* current node is not on the Dirichlet boundary */
		for (jj = 0; jj < N; jj++) {
		    j = phgSolverMapE2L(solver, 0, e, jj);
		    mat = phgQuadGradBasAGradBas(e, u_h, jj, a, u_h, ii, -1);
		    phgSolverAddMatrixEntry(solver, i, j, mat);
		}
		phgQuadDofTimesBas(e, f, u_h, ii, -1, &rhs);
	    }
	    phgSolverAddRHSEntry(solver, i, rhs);
	}
    }
}

static void
estimate_error(DOF *u_h, DOF *a, DOF *f, DOF *error)
/* compute H1 error indicator */
{
    GRID *g = u_h->g;
    ELEMENT *e;
    DOF *agrad, *jump, *residual, *tmp;

    /* tmp = \grad u_h */
    tmp = phgDofGradient(u_h, NULL, NULL, NULL);
    /* agrtad = a \grad u_h */
    agrad = phgDofMM(MAT_OP_N, MAT_OP_N, 1, 3, 1, 1., a, 1, tmp, 0., NULL);
    phgDofFree(&tmp);
    /* jump := \int_f [a\grad u_h \cdot n], one value on each face */
    jump = phgQuadFaceJump(agrad, DOF_PROJ_DOT, NULL, QUAD_DEFAULT);
    residual = phgDofDivergence(agrad, NULL, NULL, NULL);
    phgDofAXPY(1.0, f, &residual);

    ForAllElements(g, e) {
	int i;
	FLOAT eta1, eta2, h;
	/* compute sum of face jumps */
	eta1 = 0.0;
	/* for each face F compute [a\grad u_h \cdot n] */
	for (i = 0; i < NFace; i++) {
	    if (e->bound_type[i] & u_h->DB_mask)
		continue;	/* boundary face */
	    h = phgGeomGetFaceDiameter(g, e, i);	/* face diameter */
	    eta1 += *DofFaceData(jump, e->faces[i]) * h;
	}
	/* compute residual */
	h = phgGeomGetDiameter(g, e);	/* element diameter */
	eta2 = h * h * phgQuadDofDotDof(e, residual, residual, -1);
	*DofElementData(error, e->index) = eta1 * 0.5 + eta2;
    }
    phgDofFree(&jump);
    phgDofFree(&residual);
    phgDofFree(&agrad);

    return;
}

int
main(int argc, char *argv[])
{
    char *fn = "cube4.dat";	/* input mesh file */
    char *vtk = NULL;		/* VTK file (NULL -> don't write VTK file) */
    FLOAT tol = 0.0;		/* convergence criterion */
    INT mem_max = 600;		/* memory (per process) limit (MB) */
    FLOAT alpha = 1000.0;	/* the constant alpha */

    GRID *g;
    ELEMENT *e;
    DOF *u_h, *a, *f, *error;
    SOLVER *solver;
    FLOAT total_error;
    size_t mem;

    phgOptionsRegisterFilename("-mesh_file", "Mesh file (input)", (char **)&fn);
    phgOptionsRegisterFilename("-vtk_file", "VTK file (output)", (char **)&vtk);
    phgOptionsRegisterFloat("-tol", "Convergence criterion", &tol);
    phgOptionsRegisterInt("-mem_max", "Maximum memory (MB)", &mem_max);
    phgOptionsRegisterFloat("-alpha", "alpha", &alpha);

    phgOptionsPreset("-dof_type P4");

    phgInit(&argc, &argv);

    g = phgNewGrid(-1);
    if (!phgImport(g, fn, FALSE))
	phgError(1, "can't read mesh file \"%s\".\n", fn);

    /* The numerical solution, default type */
    u_h = phgDofNew(g, DOF_DEFAULT, 1, "u_h", DofInterpolation);
    phgPrintf("DOF type: %s, mesh file: %s, alpha = %lg\n",
		u_h->type->name, fn, alpha);

    /* The coefficient (material or medium), P0 type (piecewise constant) */
    a = phgDofNew(g, DOF_P0, 1, "a", DofInterpolation);
    ForAllElements(g, e) {
	int k;
	FLOAT x, y, z, barycenter[] = {0.25, 0.25, 0.25, 0.25};
	phgGeomLambda2XYZ(g, e, barycenter, &x, &y, &z);
	k = (x < 0.5 ? 1 : 0) + (y < 0.5 ? 1 : 0) + (z < 0.5 ? 1 : 0);
	*DofElementData(a, e->index) = (k % 2) ? alpha : 1.0 / alpha;
    }

    /* RHS function */
    f = phgDofNew(g, DOF_CONSTANT, 1, "f",  DofNoAction);
    phgDofSetDataByValue(f, 1.0);

    /* DOF for storing error indicators, P0 type (1 indicator per element) */
    error = phgDofNew(g, DOF_P0, 1, "error indicator", DofNoAction);

    while (TRUE) {	/* mesh adaptation loop */
	phgPrintf("*** %ld DOF, %ld elements, %d submesh%s\n",
			(long)DofGetDataCountGlobal(u_h),
			(long)g->nleaf_global,
			g->nprocs, g->nprocs <= 1 ? "" : "es");
	/* This is the only line needed for MPI parallel execution */
	if (phgBalanceGrid(g, 1.2, 1, NULL, 0.))
	    phgPrintf("    Repartition mesh, LIF: %lg\n", (double)g->lif);
	/* create solver, with DOF of u_h as its unknowns */
	solver = phgSolverCreate(SOLVER_DEFAULT, u_h, NULL);
	/* compute matrix and RHS */
	build_linear_system(solver, u_h, a, f);
	/* solve linear system */
	phgSolverSolve(solver, TRUE, u_h, NULL);
	phgPrintf("    Solve linear system, nits = %d, residual = %le\n",
			solver->nits, (double)solver->residual);
	/* destroy solver */
	phgSolverDestroy(&solver);
	/* compute error indicators */
	estimate_error(u_h, a, f, error);
	total_error = Sqrt(phgDofNormL1Vec(error));
	phgMemoryUsage(g, &mem);
	phgPrintf("    Error indicator = %0.3le, mem = %0.2lfMB\n",
			(double)total_error, (double)mem / (1024.0 * 1024.0));
	if (total_error <= tol || mem > 1024 * (size_t)mem_max * 1024)
	    break;
	/* mark elements to be refined */
	phgMarkRefine(MARK_DEFAULT, error, 0.5, NULL, 0., 1,
					Pow(tol, 2) / g->nelem_global);
	/* refine marked elements */
	phgRefineMarkedElements(g);
    }	/* while */

    if (vtk != NULL) {
	/* output a VTK file for postprocessing/visualization */
	phgPrintf("Write final solution to \"%s\".\n", vtk);
	phgExportVTK(g, vtk, u_h, a, error, NULL);
    }

    phgDofFree(&u_h);
    phgDofFree(&a);
    phgDofFree(&f);
    phgDofFree(&error);

    phgFreeGrid(&g);

    phgFinalize();

    return 0;
}
