/* 
 * This code solves the following Poisson's equation with zero Dirichlet BC:
 * 	 - \div (\grad u(x)) = 1
 */

#include "phg.h"

#include <string.h>
#include <math.h>

static void
build_linear_system(SOLVER *solver, DOF *u_h, DOF *f)
{
    GRID *g = u_h->g;
    ELEMENT *e;

    assert(u_h->dim == 1);
    ForAllElements(g, e) {	/* loop on elements */
	int N = DofGetNBas(u_h, e);	/* number of bases in the element */
	int ii, jj;
	INT i, j;
	FLOAT mat, rhs, buf[N];

	for (ii = 0; ii < N; ii++) {	/* loop on bases in current element */
	    i = phgSolverMapE2L(solver, 0, e, ii);
	    if (phgDofDirichletBC(u_h, e, ii, NULL, buf, &rhs, DOF_PROJ_NONE)) {
		/* current node is on the Dirichlet boundary */
		for (jj = 0; jj < N; jj++) {
		    j = phgSolverMapE2L(solver, 0, e, jj);
		    phgSolverAddMatrixEntry(solver, i, j, buf[jj]);
		}
	    }
	    else {
		/* current node is not on the Dirichlet boundary */
		for (jj = 0; jj < N; jj++) {
		    j = phgSolverMapE2L(solver, 0, e, jj);
		    mat = phgQuadGradBasDotGradBas(e, u_h, jj, u_h, ii, -1);
		    phgSolverAddMatrixEntry(solver, i, j, mat);
		}
		phgQuadDofTimesBas(e, f, u_h, ii, -1, &rhs);
	    }
	    phgSolverAddRHSEntry(solver, i, rhs);
	}
    }
}

int
main(int argc, char *argv[])
{
    char *fn = "cube4.dat";	/* input mesh file */
    INT refines = 0;		/* grid refinement levels */

    GRID *g;
    DOF *u_h, *f;
    SOLVER *solver;

    phgOptionsRegisterFilename("-mesh_file", "Mesh file (input)", (char **)&fn);
    phgOptionsRegisterInt("-refines", "Refinement levels", &refines);

    phgInit(&argc, &argv);

    g = phgNewGrid(-1);
    if (!phgImport(g, fn, TRUE))
	phgError(1, "can't read mesh file \"%s\".\n", fn);
    phgRefineAllElements(g, refines);

    /* The numerical solution, default type */
    u_h = phgDofNew(g, DOF_DEFAULT, 1, "u_h", DofInterpolation);
    /* RHS function */
    f = phgDofNew(g, DOF_CONSTANT, 1, "f",  DofNoAction);
    phgDofSetDataByValue(f, 1.0);

    phgPrintf("*** %ld DOF, %ld elements, %d submesh%s\n",
			(long)DofGetDataCountGlobal(u_h),
			(long)g->nleaf_global,
			g->nprocs, g->nprocs <= 1 ? "" : "es");

    /* create solver, with DOF of u_h as its unknowns */
    solver = phgSolverCreate(SOLVER_DEFAULT, u_h, NULL);
    /* compute matrix and RHS */
    build_linear_system(solver, u_h, f);
    /* solve linear system */
    phgSolverSolve(solver, TRUE, u_h, NULL);
    phgPrintf("Solve linear system, nits = %d, residual = %le\n",
			solver->nits, (double)solver->residual);
    /* destroy solver */
    phgSolverDestroy(&solver);

    phgExportVTK(g, "simple.vtk", u_h, NULL);
    phgPrintf("Solution written to \"simple.vtk\".\n");

    phgDofFree(&u_h);
    phgDofFree(&f);
    phgFreeGrid(&g);

    phgFinalize();

    return 0;
}
